from berroutils.crypter import Crypter

example_data = {"task": "encryption", "size": 42}


def test_generate_key_salt():
    """Check if key & salt are generated in the correct foramt"""
    mykey, mysalt = Crypter.generate_key_salt_from_password(password="my_password")

    assert isinstance(mykey, str)
    assert len(mykey) == 44
    assert isinstance(mysalt, str)
    assert len(mysalt) == 24


def test_en_decrypt_json(crypter):
    example_token = crypter.encrypt_json(json_data=example_data)
    result = crypter.decrypt_to_json(token=example_token)

    assert result == example_data


def test_en_decrypt(crypter):
    example_token = crypter.encrypt(data="A simple string for testing")
    result = crypter.decrypt(token=example_token)

    assert result == "A simple string for testing"


def test_decryption_from_pass_salt(crypter):
    example_token = crypter.encrypt_json(json_data=example_data)

    # password = "my_password"
    mysalt = 'jVRgK4mCjM3z3lP3dECzYg=='
    crypter_2 = Crypter.from_password_salt(password="my_password", salt=mysalt)
    result = crypter_2.decrypt_to_json(token=example_token)

    assert result == example_data



import json
import pickle
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any

from berroutils.crypter import Crypter


class FileHandler(metaclass=ABCMeta):
    def __init__(self, file_path: Path):
        self.file_path = file_path
        self.check_suffix()
        if not file_path.exists():
            self.save(data=None)

    @abstractmethod
    def load(self) -> Any:
        pass

    @abstractmethod
    def save(self, data) -> None:
        pass

    @abstractmethod
    def expected_suffix(self) -> str:
        """return the expected suffix as string"""
        pass

    def check_suffix(self):
        if self.file_path.suffix != self.expected_suffix():
            raise ValueError(
                f"File must have a {self.expected_suffix()} suffix")

    @property
    def last_modification(self):
        """Timestamp of last modification"""
        return self.file_path.stat().st_mtime


class JsonFileHandler(FileHandler):
    def load(self) -> Any:
        with open(self.file_path, "r") as file:
            return json.load(file)

    def save(self, data) -> None:
        with open(self.file_path, "w") as file:
            json.dump(data, file, indent=4)

    def expected_suffix(self) -> str:
        return ".json"


class PickleFileHandler(FileHandler):
    def load(self) -> Any:
        with open(self.file_path, "rb") as file:
            return pickle.load(file)

    def save(self, data) -> None:
        with open(self.file_path, "wb") as file:
            pickle.dump(data, file)

    def expected_suffix(self) -> str:
        return ".pickle"


class CryptoJsonFileHandler(FileHandler):
    def __init__(self, file_path: Path, crypter: Crypter):
        self.crypter = crypter
        super().__init__(file_path=file_path)

    def load(self) -> Any:
        with open(self.file_path, "rb") as file:
            token = file.read()
        return self.crypter.decrypt_to_json(token=token)

    def save(self, data) -> None:
        token = self.crypter.encrypt_json(json_data=data)
        with open(self.file_path, "wb") as file:
            file.write(token)

    def expected_suffix(self) -> str:
        return ".enc"



import base64
import json
import os
from operator import ifloordiv

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class Crypter:
    def __init__(self, key: str | bytes):
        """Initialize instance based on key as urlsafe_b64encoded string"""
        self.f = Fernet(key)

    def encrypt(self, data: str) -> bytes:
        return self.f.encrypt(data=data.encode())

    def decrypt(self, token: bytes) -> str | None:
        try:
            return self.f.decrypt(token=token).decode()
        except InvalidToken:
            print("Invalid token, most likely the key is incorrect")
            return None

    def encrypt_json(self, json_data) -> bytes:
        """Encrypt JSON to bytes."""
        json_string = json.dumps(json_data)
        return self.encrypt(data=json_string)

    def decrypt_to_json(self, token: bytes):
        """Decrypt token back to JSON"""
        decrypted = self.decrypt(token=token)
        return json.loads(decrypted) if decrypted else None

    @classmethod
    def generate_key(cls) -> bytes:
        """generate key and return it as urlsafe_b64encode encoded string"""
        return Fernet.generate_key()

    @staticmethod
    def _derive_key(salt_b: bytes, password: str) -> bytes:
        """Derive the key from the `password` using the passed `salt`

        see: https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet
        """
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt_b,
            iterations=1_200_000, )

        return kdf.derive(password.encode())

    @classmethod
    def generate_key_salt_from_password(cls, password: str, size=16) -> (str, str):
        """Generate key and salt from password and return them as base64.urlsafe_b64encode string
        """
        salt_b = os.urandom(16)  # returns bytes
        salt = cls.to_urlsafe_str(salt_b)  # convert to string
        key_b = cls._derive_key(salt_b=salt_b, password=password)
        key = cls.to_urlsafe_str(key_b)  # convert to string
        return key, salt

    @classmethod
    def from_password_salt(cls, password: str, salt: str):
        """alternative constructor to initiate instance of Crypt"""
        salt_b = cls.to_bytes(salt)  # convert to bytes
        key_b = cls._derive_key(salt_b=salt_b, password=password)
        key = cls.to_urlsafe_str(key_b)  # convert to string
        return cls(key)

    @staticmethod
    def to_urlsafe_str(raw_bytes: bytes):
        return base64.urlsafe_b64encode(raw_bytes).decode()

    @staticmethod
    def to_bytes(urlsafe_str):
        return base64.urlsafe_b64decode(urlsafe_str.encode())


if __name__ == '__main__':
    mykey, mysalt = Crypter.generate_key_salt_from_password(password="my_password")
    print(type(mykey))
    print(f"{mykey=}")
    crypter = Crypter(key=mykey)

    example_data = {"name": "Alice", "role": "admin"}

    example_token = crypter.encrypt_json(json_data=example_data)
    string_based_token = crypter.encrypt(data="A simple string for testing")
    print(f"{example_token=}")

    filename = "/Users/niko/Desktop/temp.enc"
    with open(filename, "wb") as file:
        file.write(example_token)

    # -------------

    with open(filename, "rb") as file:
        tkn = file.read()

    crypter_2 = Crypter.from_password_salt(password="my_password", salt=mysalt)

    print(f"decrypted: {crypter_2.decrypt_to_json(token=tkn)}")
    print(f"decrypted: {crypter_2.decrypt(token=string_based_token)}")



import json
import pickle

import pytest

from berroutils.crypter import Crypter
from tests import TEST_DATA


@pytest.fixture
def temp_json_file(tmp_path_factory):
    file_path = tmp_path_factory.mktemp('data') / "test.json"
    data = {"key1": 0}
    with open(file_path, "w") as f:
        json.dump(data, f)
    yield file_path


@pytest.fixture
def temp_pickle_file(tmp_path_factory):
    file_path = tmp_path_factory.mktemp('data') / "test.pickle"
    data = {"key1": 0}
    with open(file_path, "wb") as f:
        pickle.dump(data, f)
    yield file_path


@pytest.fixture
def fp_ldc_energy(tmp_path_factory):
    with open(TEST_DATA / "test_energy_meter.json") as f:  # this file will remain as is
        content = json.load(f)
    filepath = tmp_path_factory.mktemp('data') / "test_energy_tmp.json"  # this file will be modified
    with open(filepath, 'w') as f:
        json.dump(content, f)
    yield filepath


@pytest.fixture
def temp_enc_file(tmp_path_factory):
    file_path = tmp_path_factory.mktemp('data') / "test_encrypted.enc"
    return file_path


@pytest.fixture
def crypter():
    """Crypter instance

    The key has been genertated with:
    password = "my_password"
    mysalt = 'jVRgK4mCjM3z3lP3dECzYg=='
    """
    mykey = 'XgDEKKz119s_O89bHxKbMHv_RgUoJZZzYw2ZAOx6Q5M='
    return Crypter(mykey)



import json
import pickle

import pytest

from berroutils.crypter import Crypter
from berroutils.plugins.file_handler import CryptoJsonFileHandler, JsonFileHandler, PickleFileHandler, FileHandler
from tests import TEST_DATA


def test_save_json(temp_json_file):
    handler = JsonFileHandler(temp_json_file)
    data = {"key": "value"}
    handler.save(data)

    with open(temp_json_file, "r") as file:
        content = json.load(file)

    assert content == data


def test_load_json(temp_json_file):
    handler = JsonFileHandler(temp_json_file)
    content = handler.load()

    assert content == {"key1": 0}
    assert isinstance(handler, FileHandler)


def test_nonexisting_fp(tmp_path):
    """Test initiation of class with nonexiting path (handler)
    Test initiation of class with empty file (handler_2)"""
    new_path = tmp_path / "new_test.json"
    handler = JsonFileHandler(new_path)

    with open(new_path, "r") as f:
        data = json.load(f)
    assert not data

    handler_2 = JsonFileHandler(new_path)
    content = handler_2.load()
    assert not content


def test_check_suffix(temp_json_file):
    with pytest.raises(ValueError):
        JsonFileHandler(temp_json_file.with_suffix(".txt"))


def test_save_pickle(temp_pickle_file):
    handler = PickleFileHandler(temp_pickle_file)
    data = {"key": "value"}
    handler.save(data)

    with open(temp_pickle_file, "rb") as file:
        content = pickle.load(file)

    assert content == data


def test_load_pickle(temp_pickle_file):
    handler = PickleFileHandler(temp_pickle_file)
    content = handler.load()

    assert content == {"key1": 0}


def test_save_load_enc(temp_enc_file, crypter):
    handler = CryptoJsonFileHandler(file_path=temp_enc_file, crypter=crypter)
    example_data = {"task": "encryption", "size": 42}
    handler.save(data=example_data)

    with open(temp_enc_file, "rb") as file:
        content = file.read()

    assert crypter.decrypt(content) == '{"task": "encryption", "size": 42}'

    assert handler.load() == example_data


def test_load_enc_wrong_key():
    """Loading an encrypted file with the wrong key.

    Should print 'Invalid token, most likely the key is incorrect'
    and return None"""
    filepath = TEST_DATA / "test.enc"
    crypter = Crypter(key='XgDEKKz229s_O89bHxKbMHv_RgUoJZZzYw2ZAOx6Q5M=')
    handler = CryptoJsonFileHandler(file_path=filepath, crypter=crypter)

    assert handler.load() is None



import json
import pytest
from src.berroutils.container.list_of_dict_container import ListOfDictContainer, KeyCheck
from src.berroutils.plugins.file_handler import FileHandler, JsonFileHandler


class EnergyContainer(ListOfDictContainer):
    required_keys = {'date', 'gas', 'water', 'electricity'}
    alternative_keys = {"wasser": "water",
                        "strom" : "electricity", "datum": "date"}

    def __init__(self, filehandler: FileHandler):
        super().__init__(filehandler)

    def __str__(self):
        return "EnergyContainer"


@pytest.fixture
def container(fp_ldc_energy):
    json_handler = JsonFileHandler(file_path=fp_ldc_energy)
    return EnergyContainer(filehandler=json_handler)


def test_init(container):
    assert len(container) == 7
    assert container.data[0] == {
        "date"       : 1727136000000,
        "gas"        : 196.416,
        "water"      : 530.366,
        "electricity": 269.0,
        "comment"    : "Heizungswartung 2024"
    }


# @pytest.mark.skip("wip")
def test_add(container, tmp_path):
    other_data = [{
        "date"       : 1740700800000,
        "gas"        : 1904.9,
        "water"      : 625.631,
        "electricity": 2135.0,
        "comment"    : "Jahresablesung Strom"
    }]
    other_path = tmp_path / "other_test.json"
    json_handler = JsonFileHandler(file_path=other_path)
    other_container = EnergyContainer(filehandler=json_handler)
    other_container.update_from_upload(data=other_data)

    container.add(other_container)

    assert len(container) == 8
    assert container.find(search={"date": 1740700800000}) == other_data


@pytest.mark.parametrize("data, expected_result", [
    ([{
        "date"       : 1727136000000,
        "gas"        : 196.416,
        "water"      : 530.366,
        "electricity": 269.0,
        "comment"    : "Heizungswartung 2024"
    }], KeyCheck(valid=True, missing_keys=set())),
    ([{
        "datum" : 1740700800000,
        "gas"   : 1904.9,
        "wasser": 625.631,
        "strom" : 2135.0
    }], KeyCheck(valid=True, missing_keys=set())),
    ([{
        "wrong_key": 1740700800000,
        "gas"      : 1904.9,
        "strom"    : 2135.0
    }], KeyCheck(valid=False, missing_keys={"date", "water"}))])
def test_contains_required_keys(container, data, expected_result):
    key_check = container.contains_required_keys(data)
    assert key_check == expected_result


def test_update_from_file(container, tmp_path):
    new_data = [{
        "date"       : 1,
        "gas"        : 2,
        "water"      : 3,
        "electricity": 4}]
    new_path = tmp_path / "update_test.json"
    with open(new_path, "w") as f:
        json.dump(new_data, f)
    container.update_from_file(new_path)
    assert container.data == new_data


def test_get_entries(container):
    entries = container.get_entries("date", 1727136000000)
    assert entries == [{
        "date"       : 1727136000000,
        "gas"        : 196.416,
        "water"      : 530.366,
        "electricity": 269.0,
        "comment"    : "Heizungswartung 2024"
    }]


def test_find(container):
    search = {"date": 1727136000000}
    display = {"gas": 1}
    results = container.find(search, display)
    assert results == [{"gas": 196.416}]


def test_last_update(container):
    assert isinstance(container.last_update, str)



import json
import logging
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import List
from zoneinfo import ZoneInfo

import pandas as pd
from berroutils.plugins.file_handler import FileHandler

KeyCheck = namedtuple("KeyCheck", "valid missing_keys")


class ListOfDictContainer(metaclass=ABCMeta):
    required_keys: set[str] = None
    alternative_keys: dict[str] = None

    # format: alternative_keys = {"alt_key": "unique_key"}

    @abstractmethod
    def __str__(self):
        pass

    def __init__(self, filehandler: FileHandler):
        """initiate from json file"""
        self.filehandler = filehandler
        self.data = filehandler.load() or []

    def __iter__(self):
        """make the content of the class iterable"""
        return iter(self.data)

    def __len__(self):
        """return the number of entries"""
        return len(self.data)

    def add(self, other):
        """outer join with another instance of this class

        This instance will be updated with the content of the other instance"""
        if type(other) is not type(self):
            error_message = f"other is type {type(other)}, expected type {type(self)}"
            logging.warning(error_message)
            raise ValueError(error_message)
        self.data += other.data
        self.filehandler.save(data=self.data)
        return self.data

    def _clean_data(self, data):
        return data

    def _unique_keys(self, data) -> List[dict]:
        """change keys in data to predefined keys

        `alternative_keys` defines known alternative keys and maps them to the predefined keys
        All other keys will be maintained.

        Returns: data
        """
        if self.alternative_keys:
            unique = []
            for elem in data:
                unique_elem = {self.alternative_keys.get(k, k.strip("'")): v for k, v in elem.items()}
                unique.append(unique_elem)
            return unique
        else:
            return data

    def contains_required_keys(self, dataset: list[dict]) -> KeyCheck:
        """check if given dataset contains all required keys.

        Alternative keys will be recognized and converted to the corresponding required keys

        Args:
            dataset: The dataset to be checked

        Returns:
            (bool, list of missing keys)
        """
        if not self.required_keys:
            raise ValueError("required_keys are not defined")
        given_unique_keys = set(self._unique_keys(dataset[:1])[0].keys())
        valid = True if self.required_keys.issubset(given_unique_keys) else False
        missing_keys = self.required_keys - given_unique_keys

        return KeyCheck(valid, missing_keys)

    def save(self) -> None:
        self.filehandler.save(data=self.data)

    def update_from_file(self, filepath: Path) -> None:
        """update both the storage file and instance of the class from a file.
        Implemented file-types: json, csv, xlsx"""
        match filepath.suffix:
            case '.json':
                with open(filepath) as f:
                    raw_data = json.load(f)
                self.update_from_upload(data=raw_data)

            case '.csv':
                df = pd.read_csv(filepath,
                                 header=0,  # use the first row as column names
                                 encoding='utf-8')
                raw_data = df.to_dict(orient='records')
                self.update_from_upload(data=raw_data)

            case '.xlsx' | '.xls':
                df = pd.read_excel(filepath)
                raw_data = df.to_dict(orient='records')
                self.update_from_upload(data=raw_data)

            case _:
                raise NotImplementedError

    def update_from_upload(self, data: list[dict]) -> None:
        """update both the json file and instance of the classe from a dataset."""
        unique_data = self._unique_keys(data=data)
        self.data = self._clean_data(unique_data)
        self.filehandler.save(data=self.data)

    def get_entries(self, key, value, data: list[dict] = None) -> list:
        """get all entries with a given key value pair"""
        results = []
        data_to_search = data or self.data
        for entry in data_to_search:
            if str(entry.get(key)) == str(value):
                results.append(entry)
        return results

    def get_entries_partial(self, key, expression):
        """get all entries with the expression being part of the key's value"""
        results = []
        for entry in self.data:
            if expression in entry.get(key):
                results.append(entry)
        return results

    def find(self, search: dict, display: dict = None) -> List[dict]:
        """retrieve dictionary in a list of dicts that contains given key value pair(s).

        A mongo-like search function that searches data (a list of dictionaries),
        returns a list with all elements in data that fulfill the key value pairs in the search dictionary
        and if the optional display dictionary is specified, return keys with value 1 or all those not 0

        Args:
            search: key value pairs that must be contained
            display: optional, extract keys from the retrieved dictionary to be returned

        Returns:
            list of dictionaries fulfilling search and display criteria

        Example:
            find_in_list_of_dict(search: {'ID': '1EC21030423', "nombreFactura": "IMM HARMONIE CREAM 50ML"},
                                display: {"presentacionComercial": 1})
            [{"presentacionComercial": "COS-512498"}]
        """

        results = []
        for elem in (dict(line) for line in self.data):
            if all([elem.get(k, None) == v for k, v in search.items()]):
                if display:
                    if sum(display.values()) > 0:  # display has elements with 1, defining what to show
                        reduced_elem = {k: v for k, v in elem.items() if display.get(k, None) == 1}
                    else:  # display has elements with 0, thus defining what not to show
                        reduced_elem = {k: v for k, v in elem.items() if display.get(k, None) != 0}
                    results.append(reduced_elem)
                else:
                    results.append(elem)
        return results

    @property
    def last_update(self) -> str:
        """Date and time of the last change to the storage file.

        Returns:
            String in the format: 13 Feb 2025 - 15:09 (UTC)
        """
        timestamp = self.filehandler.last_modification
        d = datetime.fromtimestamp(timestamp).astimezone(ZoneInfo('UTC')).strftime('%d %b %Y - %H:%M')
        return d
